@@ -1,7 +1,7 @@
 /* io.c: i/o routines for the ed line editor */
 /* GNU ed - The GNU line editor.
-   Copyright (C) 1993, 1994 Andrew Moore, Talke Studio
-   Copyright (C) 2006-2022 Antonio Diaz Diaz.
+   Copyright (C) 1993, 1994 Andrew L. Moore, Talke Studio
+   Copyright (C) 2006-2024 Antonio Diaz Diaz.
 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -108,7 +108,8 @@ static bool trailing_escape( const char * const s, int len )
 /* If *ibufpp contains an escaped newline, get an extended line (one
    with escaped newlines) from stdin.
    The backslashes escaping the newlines are stripped.
-   Return line length in *lenp, including the trailing newline. */
+   Return line length in *lenp, including the trailing newline.
+*/
 bool get_extended_line( const char ** const ibufpp, int * const lenp,
                         const bool strip_escaped_newlines )
   {
@@ -177,7 +178,7 @@ const char * get_stdin_line( int * const sizep )
       }
     else
       {
-      buf[i++] = c; if( !c ) set_binary(); if( c != '\n' ) continue;
+      buf[i++] = c; if( c == 0 ) set_binary(); if( c != '\n' ) continue;
       ++linenum_; buf[i] = 0; *sizep = i;
       return buf;
       }
@@ -202,7 +203,7 @@ static const char * read_stream_line( const char * const filename,
     if( !resize_buffer( &buf, &bufsz, i + 2 ) ) return 0;
     c = getc( fp ); if( c == EOF ) break;
     buf[i++] = c;
-    if( !c ) set_binary();
+    if( c == 0 ) set_binary();
     else if( c == '\n' )		/* remove CR only from CR/LF pairs */
       { if( strip_cr() && i > 1 && buf[i-2] == '\r' ) { buf[i-2] = '\n'; --i; }
         break; }
@@ -227,14 +228,15 @@ static const char * read_stream_line( const char * const filename,
   }
 
 
-/* read a stream into the editor buffer;
-   return total size of data read, or -1 if error */
+/* Read a stream into the editor buffer.
+   Return number of bytes read, or -1 if error.
+*/
 static long read_stream( const char * const filename, FILE * const fp,
                          const int addr )
   {
   line_t * lp = search_line_node( addr );
   undo_t * up = 0;
-  long total_size = 0;
+  long total_size = 0;		/* number of bytes read */
   const bool o_isbinary = isbinary();
   const bool appended = ( addr == last_addr() );
   const bool o_unterminated_last_line = unterminated_last_line();
@@ -276,45 +278,38 @@ static long read_stream( const char * const filename, FILE * const fp,
 /* Read a named file/pipe into the buffer.
    Return line count, -1 if file not found, -2 if fatal error.
 */
-int read_file( const char * const filename, const int addr )
+int read_file( const char * const filename, const int addr,
+               bool * const read_onlyp )
   {
   FILE * fp;
-  long size;
   int ret;
 
   if( *filename == '!' ) fp = popen( filename + 1, "r" );
-  else
-    {
-    const char * const stripped_name = strip_escapes( filename );
-    if( !stripped_name ) return -2;
-    fp = fopen( stripped_name, "r" );
-    }
+  else if( !( fp = fopen( filename, "r+" ) ) && errno != ENOENT &&
+           ( fp = fopen( filename, "r" ) ) && read_onlyp && !modified() )
+    *read_onlyp = true;
   if( !fp )
-    {
-    show_strerror( filename, errno );
-    set_error_msg( "Cannot open input file" );
-    return -1;
-    }
-  size = read_stream( filename, fp, addr );
+    { show_strerror( filename, errno );
+      set_error_msg( "Cannot open input file" ); return -1; }
+  const long size = read_stream( filename, fp, addr );	/* file size in bytes */
   if( *filename == '!' ) ret = pclose( fp ); else ret = fclose( fp );
   if( size < 0 ) return -2;
   if( ret != 0 )
-    {
-    show_strerror( filename, errno );
-    set_error_msg( "Cannot close input file" );
-    return -2;
-    }
+    { show_strerror( filename, errno );
+      set_error_msg( "Cannot close input file" ); return -2; }
   if( !scripted() ) printf( "%lu\n", size );
-  return current_addr() - addr;
+  return current_addr() - addr;	/* current_addr updated by add_line_node */
   }
 
 
-/* write a range of lines to a stream */
+/* Write a range of lines to a stream.
+   Return number of bytes written, or -1 if error.
+*/
 static long write_stream( const char * const filename, FILE * const fp,
                           int from, const int to )
   {
   line_t * lp = search_line_node( from );
-  long size = 0;
+  long size = 0;		/* number of bytes written */
 
   while( from && from <= to )
     {
@@ -338,36 +333,26 @@ static long write_stream( const char * const filename, FILE * const fp,
   }
 
 
-/* write a range of lines to a named file/pipe; return line count */
+/* Write a range of lines to a named file/pipe.
+   Return line count, or -1 if error.
+*/
 int write_file( const char * const filename, const char * const mode,
                 const int from, const int to )
   {
   FILE * fp;
-  long size;
   int ret;
 
   if( *filename == '!' ) fp = popen( filename + 1, "w" );
-  else
-    {
-    const char * const stripped_name = strip_escapes( filename );
-    if( !stripped_name ) return -1;
-    fp = fopen( stripped_name, mode );
-    }
+  else fp = fopen( filename, mode );
   if( !fp )
-    {
-    show_strerror( filename, errno );
-    set_error_msg( "Cannot open output file" );
-    return -1;
-    }
-  size = write_stream( filename, fp, from, to );
+    { show_strerror( filename, errno );
+      set_error_msg( "Cannot open output file" ); return -1; }
+  const long size = write_stream( filename, fp, from, to );	/* bytes written */
   if( *filename == '!' ) ret = pclose( fp ); else ret = fclose( fp );
   if( size < 0 ) return -1;
   if( ret != 0 )
-    {
-    show_strerror( filename, errno );
-    set_error_msg( "Cannot close output file" );
-    return -1;
-    }
+    { show_strerror( filename, errno );
+      set_error_msg( "Cannot close output file" ); return -1; }
   if( !scripted() ) printf( "%lu\n", size );
   return ( from && from <= to ) ? to - from + 1 : 0;
   }
